"""This module provides classes that represent search objects."""

from __future__ import annotations

from typing import Literal

from pydantic import BaseModel, Field

from niconico.objects.user import UserIcon, UserRelationships

UserSearchSortKey = Literal["_personalized", "followerCount", "videoCount", "liveCount"]


class UserSearchItem(BaseModel):
    """A class that represents an item of a user search response from the NvAPI."""

    type_: Literal["userSearch"] = Field(..., alias="type")
    follower_count: int = Field(..., alias="followerCount")
    video_count: int = Field(..., alias="videoCount")
    live_count: int = Field(..., alias="liveCount")
    relationships: UserRelationships
    is_premium: bool = Field(..., alias="isPremium")
    description: str
    stripped_description: str = Field(..., alias="strippedDescription")
    short_description: str = Field(..., alias="shortDescription")
    id_: int = Field(..., alias="id")
    nickname: str
    icons: UserIcon



"""This module provides a client for watching videos on Niconico."""

from __future__ import annotations

import re
import secrets
import string
import subprocess
import time
from pathlib import Path

import requests

from niconico.base.client import BaseClient
from niconico.decorators import login_required
from niconico.exceptions import DownloadError, NicoAPIError, WatchAPIError
from niconico.objects.nvapi import AccessRightsData, NvAPIResponse
from niconico.objects.video.watch import (
    NvCommentAPIData,
    NvCommentAPIResponse,
    StoryboardResponse,
    WatchAPIData,
    WatchAPIErrorData,
    WatchAPIResponse,
    WatchData,
)


class VideoWatchClient(BaseClient):
    """A client for watching videos on Niconico."""

    def get_watch_data(self, video_id: str) -> WatchData:
        """Get the watch data of a video.

        Args:
            video_id: The ID of the video.

        Returns:
            WatchData | WatchResponseError: The watch data of the video if successful, WatchResponseError
        """
        res = self.niconico.get(f"https://www.nicovideo.jp/watch/{video_id}?responseType=json")
        if res.status_code == requests.codes.ok:
            res_cls_data = WatchAPIResponse[WatchAPIData](**res.json())
            return res_cls_data.data.response
        res_cls_error = WatchAPIResponse[WatchAPIErrorData](**res.json())
        raise WatchAPIError(response=res_cls_error.data.response)

    def generate_action_track_id(self) -> str:
        """Generate a random action track ID.

        Returns:
            str: The generated action track ID.
        """
        fh_chars = string.ascii_letters + string.digits
        fh = "".join(secrets.choice(fh_chars) for _ in range(10))
        lh = int(time.time() * 1000)
        return f"{fh}_{lh}"

    def get_outputs(self, watch_data: WatchData) -> dict[str, list[str]]:
        """Get the outputs of a video.

        Args:
            watch_data: The watch data of the video.

        Returns:
            dict[str, list[str]]: The outputs of the video.
        """
        outputs: dict[str, list[str]] = {}
        top_audio_id = None
        top_audio_quality = -1
        for audio in watch_data.media.domand.audios:
            if audio.is_available and audio.quality_level > top_audio_quality:
                top_audio_id = audio.id_
                top_audio_quality = audio.quality_level
        if top_audio_id is None:
            return outputs
        for video in watch_data.media.domand.videos:
            if video.is_available:
                outputs[video.label] = [video.id_, top_audio_id]
        return outputs

    def get_hls_content_url(self, watch_data: WatchData, outputs: list[list[str]]) -> str | None:
        """Get the HLS content URL of a video.

        Args:
            watch_data: The watch data of the video.
            outputs: The outputs. e.g.: [video_id, audio_id][]

        Returns:
            str | None: The HLS content URL of the video if successful, None otherwise.
        """
        video_id = watch_data.client.watch_id
        action_track_id = watch_data.client.watch_track_id
        access_right_key = watch_data.media.domand.access_right_key
        res = self.niconico.post(
            f"https://nvapi.nicovideo.jp/v1/watch/{video_id}/access-rights/hls?actionTrackId={action_track_id}",
            json={"outputs": outputs},
            headers={"X-Access-Right-Key": access_right_key},
        )
        if res.status_code == requests.codes.created:
            res_cls = NvAPIResponse[AccessRightsData](**res.json())
            if res_cls.data is not None:
                return res_cls.data.content_url
        return None

    @login_required(premium=True)
    def get_storyboard_url(self, watch_data: WatchData) -> str | None:
        """Get the storyboards URL of a video.

        Args:
            watch_data: The watch data of the video.

        Returns:
            str | None: The storyboards URL of the video if successful, None otherwise.
        """
        if not watch_data.media.domand.is_storyboard_available:
            return None
        video_id = watch_data.client.watch_id
        action_track_id = watch_data.client.watch_track_id
        access_right_key = watch_data.media.domand.access_right_key
        res = self.niconico.post(
            f"https://nvapi.nicovideo.jp/v1/watch/{video_id}/access-rights/storyboard?actionTrackId={action_track_id}",
            headers={"X-Access-Right-Key": access_right_key},
        )
        if res.status_code == requests.codes.created:
            res_cls = NvAPIResponse[AccessRightsData](**res.json())
            if res_cls.data is not None:
                return res_cls.data.content_url
        return None

    @login_required(premium=True)
    def download_storyboards(self, watch_data: WatchData, path: str) -> None:
        """Download the storyboards of a video.

        Args:
            watch_data: The watch data of the video.
            path: The folder path to save the storyboards.
        """
        storyboard_url = self.get_storyboard_url(watch_data)
        if storyboard_url is None:
            raise NicoAPIError(message="Failed to get the storyboards URL.")
        res = self.niconico.get(storyboard_url)
        if res.status_code == requests.codes.ok:
            res_cls = StoryboardResponse(**res.json())
            if not Path(path).is_dir():
                Path(path).mkdir(parents=True)
            if Path(f"{path}/storyboard.json").exists():
                raise DownloadError(message="The storyboard.json file already exists.")
            with Path(f"{path}/storyboard.json").open(mode="w") as f:
                f.write(res.text)
            for image in res_cls.images:
                image_res = self.niconico.get(re.sub(r"(?<=/)[^/]+(?=\?)", image.url, storyboard_url))
                if image_res.status_code == requests.codes.ok:
                    if Path(f"{path}/{image.url}.jpg").exists():
                        raise DownloadError(message=f"The storyboard image already exists: {image.url}")
                    with Path(f"{path}/{image.url}.jpg").open(mode="wb") as f:
                        f.write(image_res.content)
                else:
                    raise DownloadError(message=f"Failed to download the storyboard image: {image.url}")
        else:
            raise DownloadError(message="Failed to download the storyboards.")

    def download_video(self, watch_data: WatchData, output_label: str, path: str) -> str:
        """Download a video.

        Args:
            watch_data: The watch data of the video.
            output_label: The output label of the video.
            path: The folder path to save the video.

        Returns:
            str: The path of the downloaded video.
        """
        outputs = self.get_outputs(watch_data)
        if output_label not in outputs:
            raise DownloadError(message="The output label is not available.")
        hls_content_url = self.get_hls_content_url(watch_data, [outputs[output_label]])
        if hls_content_url is None:
            raise NicoAPIError(message="Failed to get the HLS content URL.")
        name = f"{watch_data.video.id_}_{watch_data.video.title}.mp4"
        if not Path(path).is_dir():
            Path(path).mkdir(parents=True)
        if (Path(path) / name).exists():
            raise DownloadError(message="The video file already exists.")
        cookies = {
            "domand_bid": self.niconico.session.cookies.get("domand_bid"),
        }
        commands = " ".join(
            [
                "ffmpeg",
                "-headers",
                f"'cookie: {';'.join(f'{k}={v}' for k, v in cookies.items())}'",
                "-protocol_whitelist",
                "file,http,https,tcp,tls,crypto",
                "-i",
                f"'{hls_content_url}'",
                "-c",
                "copy",
                f"'{(Path(path) / name).as_posix()}'",
            ],
        )
        try:
            with subprocess.Popen(commands, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True, text=True) as p:  # noqa: S602
                p.wait()
        except subprocess.CalledProcessError as e:
            raise DownloadError(message="Failed to download the video.") from e
        return (Path(path) / name).as_posix()

    def get_comments(self, watch_data: WatchData, *, when: int | None = None) -> NvCommentAPIData | None:
        """Get the comments of a video.

        Args:
            watch_data: The watch data of the video.
            when: The time to get the comments.

        Returns:
            list[Any]: The comments of the video.
        """
        payload = {
            "threadKey": watch_data.comment.nv_comment.thread_key,
            "params": watch_data.comment.nv_comment.params.model_dump_json(by_alias=True),
            "additionals": {},
        }
        if when is not None:
            if self.niconico.premium:
                payload["additionals"] = {"when": when}
            else:
                raise NicoAPIError(message="You must be a premium member to get the comments at a specific time.")
        res = self.niconico.post(watch_data.comment.nv_comment.server + "/v1/threads", json=payload)
        if res.status_code == requests.codes.ok:
            res_cls = NvCommentAPIResponse(**res.json())
            if res_cls.meta.status == requests.codes.ok:
                return res_cls.data
        return None



"""A module to interact with the NicoNico API."""

from __future__ import annotations

import requests

from niconico.exceptions import LoginFailureError
from niconico.video import VideoClient


class NicoNico:
    """A class to interact with the NicoNico API."""

    session: requests.Session
    logined: bool
    premium: bool

    video: VideoClient

    def __init__(self) -> None:
        """Initialize the class."""
        self.session = requests.Session()
        self.logined = False
        self.video = VideoClient(self)

    def get(self, url: str) -> requests.Response:
        """Send a GET request to a URL.

        Args:
            url (str): The URL to send the request to.

        Returns:
            requests.Response: The response object.
        """
        headers = {
            "User-Agent": "niconico.py",
            "X-Frontend-Id": "6",
            "X-Frontend-Version": "0",
        }
        return self.session.get(url, headers=headers)

    def post(self, url: str, *, json: object | None = None, headers: dict[str, str] | None = None) -> requests.Response:
        """Send a POST request to a URL.

        Args:
            url (str): The URL to send the request to.
            json (object): The data to send with the request.
            headers (dict[str, str]): The headers to send with the request.

        Returns:
            requests.Response: The response object.
        """
        req_headers = {
            "User-Agent": "niconico.py",
            "X-Frontend-Id": "6",
            "X-Frontend-Version": "0",
            "X-Niconico-Language": "ja-jp",
            "X-Request-With": "https://www.nicovideo.jp",
            "Referer": "https://www.nicovideo.jp/",
        }
        if headers is not None:
            req_headers.update(headers)
        if json is None:
            return self.session.post(url, headers=req_headers)
        return self.session.post(url, headers=req_headers, json=json)

    def login_with_mail(self, mail: str, password: str, mfa: str | None = None) -> None:
        """Login to NicoNico with a mail and password.

        Args:
            mail (str): The mail to login with.
            password (str): The password to login with.
            mfa (str | None): The MFA code to login with. Defaults to None.
        """
        self.logined = False

        res = self.session.post(
            "https://account.nicovideo.jp/login/redirector?site=niconico&next_url=%2F",
            data={
                "mail_tel": mail,
                "password": password,
                "auth_id": "1158188129",
            },
        )

        if "/login" in res.url:
            raise LoginFailureError(message="Login failed")

        if "/mfa" in res.url:
            if mfa is None:
                raise LoginFailureError(message="MFA is required")
            res = self.session.post(
                res.url,
                data={
                    "otp": mfa,
                    "device_name": "niconico.py",
                },
            )

        if res.url != "https://www.nicovideo.jp/":
            raise LoginFailureError(message="Login failed")

        if res.headers.get("x-niconico-authflag") == "1":
            self.premium = False
        elif res.headers.get("x-niconico-authflag") == "3":
            self.premium = True
        else:
            raise LoginFailureError(message="Login failed")

        self.logined = True

    def login_with_session(self, session: str) -> None:
        """Login to NicoNico with a session.

        Args:
            session (str): The session to login with.
        """
        self.logined = False

        self.session.cookies.set("user_session", session)

        res = self.session.get("https://www.nicovideo.jp/")

        if res.url != "https://www.nicovideo.jp/":
            raise LoginFailureError(message="Login failed")

        if res.headers.get("x-niconico-authflag") == "1":
            self.premium = False
        elif res.headers.get("x-niconico-authflag") == "3":
            self.premium = True
        else:
            raise LoginFailureError(message="Login failed")

        self.logined = True



"""This module provides a class that represents a user client."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import requests

from niconico.base.client import BaseClient
from niconico.decorators import login_required
from niconico.objects.nvapi import (
    NvAPIResponse,
    OwnSeriesData,
    OwnUserData,
    OwnVideosData,
    RelationshipUsersData,
    UserData,
    UserMylistsData,
    UserSeriesData,
    UserVideosData,
)
from niconico.user.search import UserSearchClient

if TYPE_CHECKING:
    from niconico.niconico import NicoNico
    from niconico.objects.user import (
        NicoUser,
        OwnNicoUser,
        UserMylistItem,
        UserSeriesItem,
        UserVideosSortKey,
        UserVideosSortOrder,
    )


class UserClient(BaseClient):
    """A class that represents a user client."""

    search: UserSearchClient

    def __init__(self, niconico: NicoNico) -> None:
        """Initialize the client."""
        super().__init__(niconico)
        self.search = UserSearchClient(niconico)

    def get_user(self, user_id: str) -> NicoUser | None:
        """Get a user by its ID.

        Args:
            user_id (str): The ID of the user.

        Returns:
            NicoUser | None: The user object if found, None otherwise.
        """
        res = self.niconico.get(f"https://nvapi.nicovideo.jp/v1/users/{user_id}")
        if res.status_code == requests.codes.ok:
            res_cls = NvAPIResponse[UserData](**res.json())
            if res_cls.data is not None:
                return res_cls.data.user
        return None

    def get_user_followers(self, user_id: str, *, page_size: int = 25, page: int = 1) -> RelationshipUsersData | None:
        """Get the followers of a user by its ID.

        Args:
            user_id (str): The ID of the user.
            page_size (int): The number of followers to get per page.
            page (int): The page number to get the followers from.

        Returns:
            RelationshipUsersData | None: The list of followers if found, None otherwise.
        """
        query = {"pageSize": str(page_size), "page": str(page)}
        query_str = "&".join([f"{key}={value}" for key, value in query.items()])
        res = self.niconico.get(f"https://nvapi.nicovideo.jp/v1/users/{user_id}/followed-by/users?{query_str}")
        if res.status_code == requests.codes.ok:
            res_cls = NvAPIResponse[RelationshipUsersData](**res.json())
            if res_cls.data is not None:
                return res_cls.data
        return None

    def get_user_followings(self, user_id: str, *, page_size: int = 25, page: int = 1) -> RelationshipUsersData | None:
        """Get the followings of a user by its ID.

        Args:
            user_id (str): The ID of the user.
            page_size (int): The number of followings to get per page.
            page (int): The page number to get the followings from.

        Returns:
            RelationshipUsersData | None: The list of followings if found, None otherwise.
        """
        query = {"pageSize": str(page_size), "page": str(page)}
        query_str = "&".join([f"{key}={value}" for key, value in query.items()])
        res = self.niconico.get(f"https://nvapi.nicovideo.jp/v1/users/{user_id}/following/users?{query_str}")
        if res.status_code == requests.codes.ok:
            res_cls = NvAPIResponse[RelationshipUsersData](**res.json())
            if res_cls.data is not None:
                return res_cls.data
        return None

    def get_user_videos(
        self,
        user_id: str,
        *,
        sort_key: UserVideosSortKey = "registeredAt",
        sort_order: UserVideosSortOrder = "asc",
        page_size: int = 30,
        page: int = 1,
        sensitive_contents: Literal["mask", "filter"] | None = None,
    ) -> UserVideosData | None:
        """Get the videos of a user by its ID.

        Args:
            user_id (str): The ID of the user.
            sort_key (UserVideosSortKey): The key to sort the videos by.
            sort_order (UserVideosSortOrder): The order to sort the videos by.
            page_size (int): The number of videos to get per page.
            page (int): The page number to get the videos from.
            sensitive_contents (Literal["mask", "filter"] | None): The sensitive contents to get.

        Returns:
            UserVideosData | None: The list of videos if found, None otherwise.
        """
        query = {
            "sortKey": sort_key,
            "sortOrder": sort_order,
            "pageSize": str(page_size),
            "page": str(page),
        }
        if sensitive_contents is not None:
            query["sensitiveContents"] = sensitive_contents
        query_str = "&".join([f"{key}={value}" for key, value in query.items()])
        res = self.niconico.get(f"https://nvapi.nicovideo.jp/v3/users/{user_id}/videos?{query_str}")
        if res.status_code == requests.codes.ok:
            res_cls = NvAPIResponse[UserVideosData](**res.json())
            if res_cls.data is not None:
                return res_cls.data
        return None

    def get_user_mylists(self, user_id: str, *, sample_item_count: int = 0) -> list[UserMylistItem]:
        """Get the mylists of a user by its ID.

        Args:
            user_id (str): The ID of the user.
            sample_item_count (int): The number of items to get from each mylist.

        Returns:
            list[UserMylistItem]: The list of mylists if found, an empty list otherwise.
        """
        query = {"sampleItemCount": str(sample_item_count)}
        query_str = "&".join([f"{key}={value}" for key, value in query.items()])
        res = self.niconico.get(f"https://nvapi.nicovideo.jp/v1/users/{user_id}/mylists?{query_str}")
        if res.status_code == requests.codes.ok:
            res_cls = NvAPIResponse[UserMylistsData](**res.json())
            if res_cls.data is not None:
                return res_cls.data.mylists
        return []

    def get_user_series(self, user_id: str, *, page_size: int = 100, page: int = 1) -> list[UserSeriesItem]:
        """Get the series of a user by its ID.

        Args:
            user_id (str): The ID of the user.
            page_size (int): The number of series to get per page.
            page (int): The page number to get the series from.

        Returns:
            list[UserSeriesData] | None: The list of series if found, None otherwise.
        """
        query = {"pageSize": str(page_size), "page": str(page)}
        query_str = "&".join([f"{key}={value}" for key, value in query.items()])
        res = self.niconico.get(f"https://nvapi.nicovideo.jp/v1/users/{user_id}/series?{query_str}")
        if res.status_code == requests.codes.ok:
            res_cls = NvAPIResponse[UserSeriesData](**res.json())
            if res_cls.data is not None:
                return res_cls.data.items
        return []

    @login_required()
    def get_own(self) -> OwnNicoUser | None:
        """Get the own user.

        Returns:
            OwnNicoUser | None: The own user object if found, None otherwise.
        """
        res = self.niconico.get("https://nvapi.nicovideo.jp/v1/users/me")
        if res.status_code == requests.codes.ok:
            res_cls = NvAPIResponse[OwnUserData](**res.json())
            if res_cls.data is not None:
                return res_cls.data.user
        return None

    @login_required()
    def get_own_followers(self, *, page_size: int = 25, page: int = 1) -> RelationshipUsersData | None:
        """Get the followers of the own user.

        Args:
            page_size (int): The number of followers to get per page.
            page (int): The page number to get the followers from.

        Returns:
            RelationshipUsersData | None: The list of followers if found, None otherwise.
        """
        query = {"pageSize": str(page_size), "page": str(page)}
        query_str = "&".join([f"{key}={value}" for key, value in query.items()])
        res = self.niconico.get(f"https://nvapi.nicovideo.jp/v1/users/me/followed-by/users?{query_str}")
        if res.status_code == requests.codes.ok:
            res_cls = NvAPIResponse[RelationshipUsersData](**res.json())
            if res_cls.data is not None:
                return res_cls.data
        return None

    @login_required()
    def get_own_followings(self, *, page_size: int = 25, page: int = 1) -> RelationshipUsersData | None:
        """Get the followings of the own user.

        Args:
            page_size (int): The number of followings to get per page.
            page (int): The page number to get the followings from.

        Returns:
            RelationshipUsersData | None: The list of followings if found, None otherwise.
        """
        query = {"pageSize": str(page_size), "page": str(page)}
        query_str = "&".join([f"{key}={value}" for key, value in query.items()])
        res = self.niconico.get(f"https://nvapi.nicovideo.jp/v1/users/me/following/users?{query_str}")
        if res.status_code == requests.codes.ok:
            res_cls = NvAPIResponse[RelationshipUsersData](**res.json())
            if res_cls.data is not None:
                return res_cls.data
        return None

    @login_required()
    def get_own_videos(
        self,
        *,
        sort_key: UserVideosSortKey = "registeredAt",
        sort_order: UserVideosSortOrder = "asc",
        page_size: int = 30,
        page: int = 1,
        sensitive_contents: Literal["mask", "filter"] | None = None,
    ) -> OwnVideosData | None:
        """Get the own videos.

        Args:
            sort_key (UserVideosSortKey): The key to sort the videos by.
            sort_order (UserVideosSortOrder): The order to sort the videos by.
            page_size (int): The number of videos to get per page.
            page (int): The page number to get the videos from.
            sensitive_contents (Literal["mask", "filter"] | None): The sensitive contents to get.

        Returns:
            OwnVideosData | None: The list of own videos if found, None otherwise.
        """
        query = {
            "sortKey": sort_key,
            "sortOrder": sort_order,
            "pageSize": str(page_size),
            "page": str(page),
        }
        if sensitive_contents is not None:
            query["sensitiveContents"] = sensitive_contents
        query_str = "&".join([f"{key}={value}" for key, value in query.items()])
        res = self.niconico.get(f"https://nvapi.nicovideo.jp/v2/users/me/videos?{query_str}")
        if res.status_code == requests.codes.ok:
            res_cls = NvAPIResponse[OwnVideosData](**res.json())
            if res_cls.data is not None:
                return res_cls.data
        return None

    @login_required()
    def get_own_mylists(self, *, sample_item_count: int = 0) -> list[UserMylistItem]:
        """Get the own mylists.

        Args:
            sample_item_count (int): The number of items to get from each mylist.

        Returns:
            list[UserMylistItem]: The list of own mylists if found, an empty list otherwise.
        """
        query = {"sampleItemCount": str(sample_item_count)}
        query_str = "&".join([f"{key}={value}" for key, value in query.items()])
        res = self.niconico.get(f"https://nvapi.nicovideo.jp/v1/users/me/mylists?{query_str}")
        if res.status_code == requests.codes.ok:
            res_cls = NvAPIResponse[UserMylistsData](**res.json())
            if res_cls.data is not None:
                return res_cls.data.mylists
        return []

    @login_required()
    def get_own_series(self, *, page_size: int = 100, page: int = 1) -> list[UserSeriesItem]:
        """Get the series of a user by its ID.

        Args:
            user_id (str): The ID of the user.
            page_size (int): The number of series to get per page.
            page (int): The page number to get the series from.

        Returns:
            list[UserSeriesData] | None: The list of series if found, None otherwise.
        """
        query = {"pageSize": str(page_size), "page": str(page)}
        query_str = "&".join([f"{key}={value}" for key, value in query.items()])
        res = self.niconico.get(f"https://nvapi.nicovideo.jp/v1/users/me/series?{query_str}")
        if res.status_code == requests.codes.ok:
            res_cls = NvAPIResponse[OwnSeriesData](**res.json())
            if res_cls.data is not None:
                return res_cls.data.items
        return []



"""Module for user search client."""

from __future__ import annotations

from typing import TYPE_CHECKING

import requests

from niconico.base.client import BaseClient
from niconico.objects.nvapi import NvAPIResponse, UserSearchData

if TYPE_CHECKING:
    from niconico.objects.user.search import UserSearchSortKey


class UserSearchClient(BaseClient):
    """Client for user search."""

    def search_users(
        self,
        keyword: str,
        *,
        sort_key: UserSearchSortKey = "_personalized",
        page_size: int = 100,
        page: int = 1,
    ) -> UserSearchData | None:
        """Search users by keyword.

        Args:
            keyword (str): Keyword to search.
            sort_key (UserSearchSortKey, optional): Sort key. Defaults to "_personalized".
            page_size (int, optional): Page size. Defaults to 100.
            page (int, optional): Page. Defaults to 1.

        Returns:
            UserSearchData | None: User search data.
        """
        query = {
            "keyword": keyword,
            "sortKey": sort_key,
            "pageSize": page_size,
            "page": page,
        }
        query_str = "&".join(f"{key}={value}" for key, value in query.items())
        res = self.niconico.get(f"https://nvapi.nicovideo.jp/v1/search/user?{query_str}")
        if res.status_code == requests.codes.ok:
            res_cls = NvAPIResponse[UserSearchData](**res.json())
            if res_cls.data is not None:
                return res_cls.data
        return None



"""This module provides the video search client."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import requests

from niconico.base.client import BaseClient
from niconico.objects.nvapi import FacetData, ListSearchData, NvAPIResponse, VideoSearchData

if TYPE_CHECKING:
    from niconico.objects.video.search import (
        FacetItem,
        ListSearchSortKey,
        ListType,
        VideoSearchSortKey,
        VideoSearchSortOrder,
    )


class VideoSearchClient(BaseClient):
    """A class that represents a video search client."""

    def search_videos_by_keyword(
        self,
        keyword: str,
        *,
        sort_key: VideoSearchSortKey = "hot",
        sort_order: VideoSearchSortOrder = "none",
        page_size: int = 25,
        page: int = 1,
        sensitive_content: Literal["mask", "filter"] | None = None,
        channel_video_listing_status: Literal["included"] | None = None,
        allow_future_contents: bool | None = None,
        search_by_user: bool | None = None,
        min_registered_at: str | None = None,
        max_registered_at: str | None = None,
        max_duration: int | None = None,
    ) -> VideoSearchData | None:
        """Search videos by a keyword.

        Args:
            keyword (str): The keyword to search.
            sort_key (VideoSearchSortKey): The sort key.
            sort_order (VideoSearchSortOrder): The sort order.
            page_size (int): The page size.
            page (int): The page.
            sensitive_content (Literal["mask", "filter"] | None): The sensitive content.
            channel_video_listing_status (Literal["included"] | None): The channel video listing status.
            allow_future_contents (bool | None): The allow future contents.
            search_by_user (bool | None): The search by user.
            min_registered_at (str | None): The minimum registered at.
            max_registered_at (str | None): The maximum registered at.
            max_duration (int | None): The maximum duration.

        Returns:
            VideoSearchData | None: The search result.
        """
        query = {
            "keyword": keyword,
            "sortKey": sort_key,
            "sortOrder": sort_order,
            "pageSize": str(page_size),
            "page": str(page),
        }
        if sensitive_content is not None:
            query["sensitiveContent"] = sensitive_content
        if channel_video_listing_status is not None:
            query["channelVideoListingStatus"] = channel_video_listing_status
        if allow_future_contents is not None:
            query["allowFutureContents"] = "true" if allow_future_contents else "false"
        if search_by_user is not None:
            query["searchByUser"] = "true" if search_by_user else "false"
        if min_registered_at is not None:
            query["minRegisteredAt"] = min_registered_at
        if max_registered_at is not None:
            query["maxRegisteredAt"] = max_registered_at
        if max_duration is not None:
            query["maxDuration"] = str(max_duration)
        query_str = "&".join([f"{key}={value}" for key, value in query.items()])
        res = self.niconico.get(f"https://nvapi.nicovideo.jp/v2/search/video?{query_str}")
        if res.status_code == requests.codes.ok:
            res_cls = NvAPIResponse[VideoSearchData](**res.json())
            if res_cls.data is not None:
                return res_cls.data
        return None

    def search_videos_by_tag(
        self,
        tag: str,
        *,
        sort_key: VideoSearchSortKey = "hot",
        sort_order: VideoSearchSortOrder = "none",
        page_size: int = 25,
        page: int = 1,
        sensitive_content: Literal["mask", "filter"] | None = None,
        channel_video_listing_status: Literal["included"] | None = None,
        allow_future_contents: bool | None = None,
        search_by_user: bool | None = None,
        min_registered_at: str | None = None,
        max_registered_at: str | None = None,
        max_duration: int | None = None,
    ) -> VideoSearchData | None:
        """Search videos by a tag.

        Args:
            tag (str): The tag to search.
            sort_key (VideoSearchSortKey): The sort key.
            sort_order (VideoSearchSortOrder): The sort order.
            page_size (int): The page size.
            page (int): The page.
            sensitive_content (Literal["mask", "filter"] | None): The sensitive content.
            channel_video_listing_status (Literal["included"] | None): The channel video listing status.
            allow_future_contents (bool | None): The allow future contents.
            search_by_user (bool | None): The search by user.
            min_registered_at (str | None): The minimum registered at.
            max_registered_at (str | None): The maximum registered at.
            max_duration (int | None): The maximum duration.

        Returns:
            VideoSearchData | None: The search result.
        """
        query = {
            "tag": tag,
            "sortKey": sort_key,
            "sortOrder": sort_order,
            "pageSize": str(page_size),
            "page": str(page),
        }
        if sensitive_content is not None:
            query["sensitiveContent"] = sensitive_content
        if channel_video_listing_status is not None:
            query["channelVideoListingStatus"] = channel_video_listing_status
        if allow_future_contents is not None:
            query["allowFutureContents"] = "true" if allow_future_contents else "false"
        if search_by_user is not None:
            query["searchByUser"] = "true" if search_by_user else "false"
        if min_registered_at is not None:
            query["minRegisteredAt"] = min_registered_at
        if max_registered_at is not None:
            query["maxRegisteredAt"] = max_registered_at
        if max_duration is not None:
            query["maxDuration"] = str(max_duration)
        query_str = "&".join([f"{key}={value}" for key, value in query.items()])
        res = self.niconico.get(f"https://nvapi.nicovideo.jp/v2/search/video?{query_str}")
        if res.status_code == requests.codes.ok:
            res_cls = NvAPIResponse[VideoSearchData](**res.json())
            if res_cls.data is not None:
                return res_cls.data
        return None

    def get_facet_by_keyword(
        self,
        keyword: str,
        *,
        sort_key: VideoSearchSortKey = "hot",
        sort_order: VideoSearchSortOrder = "none",
        sensitive_content: Literal["mask", "filter"] | None = None,
        channel_video_listing_status: Literal["included"] | None = None,
        allow_future_contents: bool | None = None,
        search_by_user: bool | None = None,
        min_registered_at: str | None = None,
        max_registered_at: str | None = None,
        max_duration: int | None = None,
    ) -> list[FacetItem]:
        """Get the number of videos for each genre of videos searched with specified conditions.

        Args:
            keyword (str): The keyword to search.
            sort_key (VideoSearchSortKey): The sort key.
            sort_order (VideoSearchSortOrder): The sort order.
            sensitive_content (Literal["mask", "filter"] | None): The sensitive content.
            channel_video_listing_status (Literal["included"] | None): The channel video listing status.
            allow_future_contents (bool | None): The allow future contents.
            search_by_user (bool | None): The search by user.
            min_registered_at (str | None): The minimum registered at.
            max_registered_at (str | None): The maximum registered at.
            max_duration (int | None): The maximum duration.

        Returns:
            list[FacetItem]: The facet items.
        """
        query = {
            "keyword": keyword,
            "sortKey": sort_key,
            "sortOrder": sort_order,
        }
        if sensitive_content is not None:
            query["sensitiveContent"] = sensitive_content
        if channel_video_listing_status is not None:
            query["channelVideoListingStatus"] = channel_video_listing_status
        if allow_future_contents is not None:
            query["allowFutureContents"] = "true" if allow_future_contents else "false"
        if search_by_user is not None:
            query["searchByUser"] = "true" if search_by_user else "false"
        if min_registered_at is not None:
            query["minRegisteredAt"] = min_registered_at
        if max_registered_at is not None:
            query["maxRegisteredAt"] = max_registered_at
        if max_duration is not None:
            query["maxDuration"] = str(max_duration)
        query_str = "&".join([f"{key}={value}" for key, value in query.items()])
        res = self.niconico.get(f"https://nvapi.nicovideo.jp/v2/search/facet?{query_str}")
        if res.status_code == requests.codes.ok:
            res_cls = NvAPIResponse[FacetData](**res.json())
            if res_cls.data is not None:
                return res_cls.data.items
        return []

    def search_facet_by_tag(
        self,
        tag: str,
        *,
        sort_key: VideoSearchSortKey = "hot",
        sort_order: VideoSearchSortOrder = "none",
        sensitive_content: Literal["mask", "filter"] | None = None,
        channel_video_listing_status: Literal["included"] | None = None,
        allow_future_contents: bool | None = None,
        search_by_user: bool | None = None,
        min_registered_at: str | None = None,
        max_registered_at: str | None = None,
        max_duration: int | None = None,
    ) -> list[FacetItem]:
        """Search videos by a tag.

        Args:
            tag (str): The tag to search.
            sort_key (VideoSearchSortKey): The sort key.
            sort_order (VideoSearchSortOrder): The sort order.
            sensitive_content (Literal["mask", "filter"] | None): The sensitive content.
            channel_video_listing_status (Literal["included"] | None): The channel video listing status.
            allow_future_contents (bool | None): The allow future contents.
            search_by_user (bool | None): The search by user.
            min_registered_at (str | None): The minimum registered at.
            max_registered_at (str | None): The maximum registered at.
            max_duration (int | None): The maximum duration.

        Returns:
            list[FacetItem]: The facet items.
        """
        query = {
            "tag": tag,
            "sortKey": sort_key,
            "sortOrder": sort_order,
        }
        if sensitive_content is not None:
            query["sensitiveContent"] = sensitive_content
        if channel_video_listing_status is not None:
            query["channelVideoListingStatus"] = channel_video_listing_status
        if allow_future_contents is not None:
            query["allowFutureContents"] = "true" if allow_future_contents else "false"
        if search_by_user is not None:
            query["searchByUser"] = "true" if search_by_user else "false"
        if min_registered_at is not None:
            query["minRegisteredAt"] = min_registered_at
        if max_registered_at is not None:
            query["maxRegisteredAt"] = max_registered_at
        if max_duration is not None:
            query["maxDuration"] = str(max_duration)
        query_str = "&".join([f"{key}={value}" for key, value in query.items()])
        res = self.niconico.get(f"https://nvapi.nicovideo.jp/v2/search/facet?{query_str}")
        if res.status_code == requests.codes.ok:
            res_cls = NvAPIResponse[FacetData](**res.json())
            if res_cls.data is not None:
                return res_cls.data.items
        return []

    def search_lists(
        self,
        keyword: str,
        sort_key: ListSearchSortKey = "_hotTotalScore",
        types: list[ListType] | None = None,
        page_size: int = 100,
        page: int = 1,
    ) -> ListSearchData | None:
        """Search lists.

        Args:
            keyword (str): The keyword to search.
            sort_key (ListSearchSortKey): The sort key.
            types (list[ListType]): The types. If None, all types are included.
            page_size (int): The page size.
            page (int): The page.

        Returns:
            ListSearchData | None: The search result.
        """
        query = {
            "keyword": keyword,
            "sortKey": sort_key,
            "pageSize": str(page_size),
            "page": str(page),
        }
        if types is not None and len(types) == 1:
            query["type"] = types[0]
        query_str = "&".join([f"{key}={value}" for key, value in query.items()])
        res = self.niconico.get(f"https://nvapi.nicovideo.jp/v1/search/list?{query_str}")
        if res.status_code == requests.codes.ok:
            res_cls = NvAPIResponse[ListSearchData](**res.json())
            if res_cls.data is not None:
                return res_cls.data
        return None



"""This module contains exceptions for the niconico package."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from niconico.objects.video.watch import WatchResponseError


class LoginFailureError(Exception):
    """An exception raised when login fails."""

    def __init__(self, message: str) -> None:
        """Initialize the exception."""
        self.message = message
        super().__init__(message)


class DownloadError(Exception):
    """An exception raised when download fails."""

    def __init__(self, message: str) -> None:
        """Initialize the exception."""
        self.message = message
        super().__init__(message)


class NicoAPIError(Exception):
    """An exception raised when an error occurs in the Nico API."""

    def __init__(self, message: str) -> None:
        """Initialize the exception."""
        self.message = message
        super().__init__(message)


class LoginRequiredError(Exception):
    """An exception raised when login is required."""

    def __init__(self, message: str) -> None:
        """Initialize the exception."""
        self.message = message
        super().__init__(message)


class PremiumRequiredError(Exception):
    """An exception raised when a premium account is required."""

    def __init__(self, message: str) -> None:
        """Initialize the exception."""
        self.message = message
        super().__init__(message)


class WatchAPIError(Exception):
    """An exception raised when an error occurs in the watch API."""

    def __init__(self, response: WatchResponseError) -> None:
        """Initialize the exception."""
        self.response = response
        super().__init__(response.error_code)



"""This module contains the class that represents a ranking object."""

from __future__ import annotations

from pydantic import BaseModel


class Genre(BaseModel):
    """A class that represents a genre."""

    key: str
    label: str



"""Main module for niconico package."""

import argparse
import logging
from pathlib import Path

from niconico import NicoNico, __version__
from niconico.exceptions import LoginFailureError, WatchAPIError


def command_details(args: argparse.Namespace) -> None:
    """Show video details."""
    logging.debug("Starting command: details with args: %s", args)
    client = NicoNico()
    video = client.video.get_video(args.id)
    if video is None:
        logging.error("Video not found")
        return
    logging.debug("Video: %s", video)
    if args.output is None:
        logging.info(video.model_dump_json(indent=2, by_alias=True))
    else:
        with args.output.open("w") as f:
            f.write(video.model_dump_json(by_alias=True))
        logging.info("Saved to %s", args.output)


def command_quality(args: argparse.Namespace) -> None:
    """Show video quality."""
    logging.debug("Starting command: quality with args: %s", args)
    client = NicoNico()
    if args.session is not None:
        try:
            client.login_with_session(args.session)
            logging.debug("Logged in with session")
        except LoginFailureError:
            logging.exception("Login failed")
        logging.debug("is_premium: %s", client.premium)
    try:
        watch_data = client.video.watch.get_watch_data(args.id)
    except WatchAPIError:
        logging.exception("An error has occurred")
        return
    logging.debug("Watch data: %s", watch_data)
    outputs = client.video.watch.get_outputs(watch_data)
    logging.info("Label\tVideo, Audio")
    for output_key, output in outputs.items():
        logging.info("%s:\t%s", output_key, ", ".join(output))


def command_download(args: argparse.Namespace) -> None:
    """Download video."""
    logging.debug("Starting command: download with args: %s", args)
    client = NicoNico()
    if args.session is not None:
        try:
            client.login_with_session(args.session)
            logging.debug("Logged in with session")
        except LoginFailureError:
            logging.exception("Login failed")
        logging.debug("is_premium: %s", client.premium)
    try:
        watch_data = client.video.watch.get_watch_data(args.id)
    except WatchAPIError:
        logging.exception("An error has occurred")
        return
    logging.debug("Watch data: %s", watch_data)
    outputs = client.video.watch.get_outputs(watch_data)
    logging.debug("Outputs: %s", outputs)
    if args.quality is not None:
        if args.quality not in outputs:
            logging.error("This quality is not supported")
            return
        quality = args.quality
        logging.debug("Selected quality: %s", quality)
    else:
        quality = next(iter(outputs))
        logging.debug("Selected quality: %s(best)", quality)
    logging.debug("Downloading video...")
    downloaded_path = client.video.watch.download_video(watch_data, quality, args.output)
    logging.info("Downloaded to %s", downloaded_path)


def main() -> None:
    """Main function for niconico package."""
    parser = argparse.ArgumentParser(
        description="download video from niconico",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("-v", "--version", action="version", version=f"%(prog)s {__version__}")
    parser.add_argument("--debug", help="enable debug logging", action="store_true")

    subparsers = parser.add_subparsers(title="commands")

    parser_details = subparsers.add_parser("details", help="show video details")
    parser_details.add_argument("id", help="video id", type=str)
    parser_details.add_argument("-o", "--output", help="output file (default: stdout)", default=None, type=Path)
    parser_details.set_defaults(func=command_details)

    parser_quality = subparsers.add_parser("quality", help="show video quality")
    parser_quality.add_argument("id", help="video id", type=str)
    parser_quality.add_argument("-s", "--session", help="user_session cookie", default=None, type=str)
    parser_quality.set_defaults(func=command_quality)

    parser_download = subparsers.add_parser("download", help="download video")
    parser_download.add_argument("id", help="video id", type=str)
    parser_download.add_argument("-o", "--output", help="output file (default: stdout)", default=".", type=Path)
    parser_download.add_argument("-s", "--session", help="user_session cookie", default=None, type=str)
    parser_download.add_argument("-q", "--quality", help="video quality (default: best)", default=None, type=str)
    parser_download.set_defaults(func=command_download)

    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO, format="%(message)s")

    if hasattr(args, "func"):
        args.func(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()
